/*****************************************************************************
 *                                                                            *
 *  -------------------------------- list.c --------------------------------  *
 *                                                                            *
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "list.h"

/*****************************************************************************
 *                                                                            *
 *  ------------------------------- list_init ------------------------------  *
 *                                                                            *
 *****************************************************************************/

void list_init(List *list, void (*destroy)(void *data)) {
  
    /*****************************************************************************
     *                                                                            *
     *  Initialize the list.                                                      *
     *                                                                            *
     *****************************************************************************/
    
    list->size = 0;
    list->destroy = destroy;
    list->head = NULL;
    list->tail = NULL;
    
    return;
    
}

/*****************************************************************************
 *                                                                            *
 *  ----------------------------- list_destroy -----------------------------  *
 *                                                                            *
 *****************************************************************************/

void list_destroy(List *list) {
  
    void               *data;
    
    /*****************************************************************************
     *                                                                            *
     *  Remove each element.                                                      *
     *                                                                            *
     *****************************************************************************/
    
    while (list_size(list) > 0) {
      
	if (list_rem_next(list, NULL, (void **)&data) == 0 && list->destroy !=
	    NULL) {
	  
	    /***********************************************************************
	     *                                                                      *
	     *  Call a user-defined function to free dynamically allocated data.    *
	     *                                                                      *
	     ***********************************************************************/
	    
	    list->destroy(data);
	    
	}
      
    }
  
    /*****************************************************************************
     *                                                                            *
     *  No operations are allowed now, but clear the structure as a precaution.   *
     *                                                                            *
     *****************************************************************************/

    memset(list, 0, sizeof(List));

    return;
    
}





/*****************************************************************************
 *                                                                            *
 *  ----------------------------- list_ins_next ----------------------------  *
 *                                                                            *
 *****************************************************************************/

int list_ins_next(List *list, ListElmt *element, const void *data) {
  
    ListElmt           *new_element;
    //printf("data avant:%i \n",element);
    /*****************************************************************************
     *                                                                            *
     *  Allocate storage for the element.                                         *
     *                                                                            *
     *****************************************************************************/

    if ((new_element = (ListElmt *)malloc(sizeof(ListElmt))) == NULL)
      return -1;

	/*****************************************************************************
	 *                                                                            *
	 *  Insert the element into the list.                                         *
	 *                                                                            *
	 *****************************************************************************/

	new_element->data = (void *)data;
	if (element == NULL) {

	    /**************************************************************************
	     *                                                                         *
	     *  Handle insertion at the head of the list.                              *
	     *                                                                         *
	     **************************************************************************/

	    if (list_size(list) == 0)
	      list->tail = new_element;

		new_element->next = list->head;
		list->head = new_element;
	}
	else {

	  /**************************************************************************
	   *                                                                         *
	   *  Handle insertion somewhere other than at the head.                     *
	   *                                                                         *
	   **************************************************************************/

	  if (element->next == NULL)
	    list->tail = new_element;

	      new_element->next = element->next;
	      element->next = new_element;

	}

    /*****************************************************************************
     *                                                                            *
     *  Adjust the size of the list to account for the inserted element.          *
     *                                                                            *
     *****************************************************************************/

    list->size++;
    return 0;
}

/*****************************************************************************
 *                                                                            *
 *  ----------------------------- list_rem_next ----------------------------  *
 *                                                                            *
 *****************************************************************************/

int list_rem_next(List *list, ListElmt *element, void **data) {

  ListElmt           *old_element;

    /*****************************************************************************
     *                                                                            *
     *  Do not allow removal from an empty list.                                  *
     *                                                                            *
     *****************************************************************************/

    if (list_size(list) == 0)
      return -1;

	/*****************************************************************************
	 *                                                                            *
	 *  Remove the element from the list.                                         *
	 *                                                                            *
	 *****************************************************************************/

	if (element == NULL) {

	    /**************************************************************************
	     *                                                                         *
	     *  Handle removal from the head of the list.                              *
	     *                                                                         *
	     **************************************************************************/

	    *data = list->head->data;
	    old_element = list->head;
	    list->head = list->head->next;

	    if (list_size(list) == 1)
	      list->tail = NULL;
	}

	else {

	    /**************************************************************************
	     *                                                                         *
	     *  Handle removal from somewhere other than the head.                     *
	     *                                                                         *
	     **************************************************************************/

	    if (element->next == NULL)
	      return -1;

		*data = element->next->data;
		old_element = element->next;
		element->next = element->next->next;

		if (element->next == NULL)
		  list->tail = element;

	}

    /*****************************************************************************
     *                                                                            *
     *  Free the storage allocated by the abstract data type.                     *
     *                                                                            *
     *****************************************************************************/

    free(old_element);

    /*****************************************************************************
     *                                                                            *
     *  Adjust the size of the list to account for the removed element.           *
     *                                                                            *
     *****************************************************************************/
    list->size--;
    return 0;
}


/*****************************************************************************
 *                                                                            *
 *  ----------------------------- list_prev -----------------------------  *
 *                                                                            *
 *****************************************************************************/

ListElmt *list_prev(List *list, ListElmt *element) {
  ListElmt           *elmt;
    /*****************************************************************************
     *                                                                            *
     *  Find the element before a given element                                                      *
     *                                                                            *
     *****************************************************************************/
    elmt = list_head(list);
  if(elmt==NULL) return NULL;

    while(elmt!=NULL) {
      if(elmt->next != NULL && elmt->next==element) return elmt;
      elmt = list_next(elmt);
    }
    return NULL;
}





int list_rem(List *list, ListElmt *element) {

  ListElmt           *prev_element;

  if (list_size(list) == 0)
    return -1;

    if (element == list->head) {

	prev_element = list->head;
	list->head = list->head->next;

	if (list_size(list) == 1)
	  list->tail = NULL;

    }
    else {

	prev_element = list_prev(list,element);
	if(prev_element==NULL) return -1;
	  prev_element->next = element->next;

	    if(list->tail==element){
	      list->tail = prev_element;
	    }
    }
    free(element);
    list->size--;
    return 0;
}






List *list_sort(List *list,int (* cmp_func)(void *p1, void *p2))
{
  ListElmt *remain_to_sort;
  ListElmt *list_ptr;
  ListElmt *list_smallest;

  int i;
  int nb_already_sorted=0;
  long int size=list->size;

  while(nb_already_sorted<size){

    remain_to_sort=list_head(list);
    for(i=0;i<nb_already_sorted;i++) remain_to_sort=list_next(remain_to_sort);

    list_ptr=remain_to_sort;
    list_smallest=list_ptr;
    while(list_ptr!=NULL){
      if(cmp_func(list_data(list_ptr),list_data(list_smallest)) < 0)
      {
	list_smallest=list_ptr;
      }

      list_ptr=list_next(list_ptr);
    }

    list_swap(list,list_smallest,remain_to_sort);

    nb_already_sorted++;
  }

  return list;
}



void list_swap(List *list, ListElmt *e1, ListElmt *e2) {
  ListElmt *tmp;

  tmp=e1->data;
  e1->data=e2->data;
  e2->data=tmp;


}



ListElmt *list_find_data_next(ListElmt *start, void *data)
{
  ListElmt* tmp=start;

  while(1)
  {
    if(tmp!=NULL && list_data(tmp)==data){
      return tmp;
    }
    else{
      if(list_is_tail(tmp)) return NULL;
      tmp=list_next(tmp);
    }
  }
  return NULL;

}




ListElmt *list_match_next(ListElmt *start, void *data, int (* cmp_func)(void *p1, void *p2))
{
  ListElmt* tmp=start;

  while(1)
  {
    if(tmp!=NULL && cmp_func(list_data(tmp),data)){
      return tmp;
    }
    else{
      if(list_is_tail(tmp)) return NULL;
      tmp=list_next(tmp);
    }
  }
  return NULL;

}
